"""Asynchronous Python client providing Urban Data information of Hamburg."""


class UDPHamburgError(Exception):
    """Generic Urban Data Platform Hamburg exception."""


class UDPHamburgConnectionError(UDPHamburgError):
    """Urban Data Platform Hamburg - connection error."""



"""Asynchronous Python client providing Urban Data information of Hamburg."""

from .exceptions import UDPHamburgConnectionError, UDPHamburgError
from .hamburg import UDPHamburg
from .models import DisabledParking, Garage, ParkAndRide

__all__ = [
    "UDPHamburg",
    "UDPHamburgConnectionError",
    "UDPHamburgError",
    "DisabledParking",
    "ParkAndRide",
    "Garage",
]



"""Asynchronous Python client providing Urban Data information of Hamburg."""
from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import Any

import pytz


@dataclass
class DisabledParking:
    """Object representing a disabled parking."""

    spot_id: str
    street: str | None
    limitation: str | None
    number: int
    longitude: float
    latitude: float

    @classmethod
    def from_dict(cls: type[DisabledParking], data: dict[str, Any]) -> DisabledParking:
        """Return a DisabledParking object from a dictionary.

        Args:
        ----
            data: The data from the API.

        Returns:
        -------
            A DisabledParking object.
        """

        def strip_spaces(string: str) -> str | None:
            """Strip spaces from a string.

            Args:
            ----
                string: The string to strip.

            Returns:
            -------
                The string without spaces or None if the string is empty.
            """
            if string is None:
                return None
            return string.strip()

        attr = data["properties"]
        geo = data["geometry"]["coordinates"]
        return cls(
            spot_id=str(data.get("id")),
            street=strip_spaces(attr.get("nahe_adresse")),
            limitation=strip_spaces(attr.get("befristung")),
            number=attr.get("anzahl"),
            longitude=geo[0],
            latitude=geo[1],
        )


@dataclass
class ParkAndRide:
    """Object representing a park and ride spot."""

    spot_id: str
    name: str
    park_type: str
    address: str
    construction_year: int
    public_transport_line: str
    disabled_parking_spaces: int
    tickets: dict[str, int]
    url: str

    free_space: int
    capacity: int
    availability_pct: float | None

    longitude: float
    latitude: float
    updated_at: datetime

    @classmethod
    def from_dict(cls: type[ParkAndRide], data: dict[str, Any]) -> ParkAndRide:
        """Return a ParkAndRide object from a dictionary.

        Args:
        ----
            data: The data from the API.

        Returns:
        -------
            A ParkAndRide object.
        """
        attr = data["properties"]
        geo = data["geometry"]["coordinates"]
        return cls(
            spot_id=str(data.get("id")),
            name=attr.get("name"),
            park_type=attr.get("art"),
            address=attr.get("adresse"),
            construction_year=attr.get("baujahr"),
            public_transport_line=attr.get("linie"),
            disabled_parking_spaces=int(attr.get("stellplaetze_behinderte_gesamt")),
            tickets={
                "day": attr.get("ticket_1_tag"),
                "month": attr.get("ticket_30_tage"),
                "year": attr.get("ticket_1_jahr"),
            },
            url=attr.get("homepage"),
            free_space=int(attr.get("stellplaetze_frei")),
            capacity=int(attr.get("stellplaetze_gesamt")),
            availability_pct=availability_calc(
                attr.get("stellplaetze_frei"),
                attr.get("stellplaetze_gesamt"),
            ),
            longitude=geo[0],
            latitude=geo[1],
            updated_at=datetime.strptime(
                attr.get("aktualitaet_belegungsdaten"),
                "%Y-%m-%d %H:%M:%S",
            ).astimezone(pytz.timezone("Europe/Berlin")),
        )


@dataclass
class Garage:
    """Object representing a garage."""

    spot_id: str
    name: str
    park_type: str
    disabled_parking_spaces: int | None
    status: str
    address: str
    price: str
    data_origin: str | None

    free_space: int | None
    capacity: int | None
    availability_pct: float | None

    longitude: float
    latitude: float
    updated_at: datetime | None

    @classmethod
    def from_dict(cls: type[Garage], data: dict[str, Any]) -> Garage:
        """Return a Garage object from a dictionary.

        Args:
        ----
            data: The data from the API.

        Returns:
        -------
            A Garage object.
        """
        attr = data["properties"]
        geo = data["geometry"]["coordinates"]
        return cls(
            spot_id=str(data.get("id")),
            name=attr.get("name"),
            park_type=attr.get("art"),
            disabled_parking_spaces=attr.get("behindertenst"),
            status=attr.get("situation"),
            address=f"{attr.get('strasse')} {attr.get('hausnr')}",
            price=attr.get("preise"),
            data_origin=attr.get("datenherkunft"),
            free_space=attr.get("frei"),
            capacity=attr.get("stellplaetze_gesamt"),
            availability_pct=availability_calc(
                attr.get("frei"),
                attr.get("stellplaetze_gesamt"),
            ),
            longitude=geo[0],
            latitude=geo[1],
            updated_at=strptime(attr.get("received"), "%d.%m.%Y, %H:%M"),
        )


def availability_calc(
    free_space: int,
    capacity: int,
    default: None = None,
) -> float | None:
    """Calculate the availability percentage.

    Args:
    ----
        free_space: The free space.
        capacity: The capacity.
        default: The default value.

    Returns:
    -------
        The availability percentage.
    """
    try:
        return round(
            (float(free_space) / float(capacity)) * 100,
            1,
        )
    except TypeError:
        return default
    except ZeroDivisionError:
        return None


def strptime(date_string: str, date_format: str, default: None = None) -> Any:
    """Strptime function with default value.

    Args:
    ----
        date_string: The date string.
        date_format: The format of the date string.
        default: The default value.

    Returns:
    -------
        The datetime object.
    """
    try:
        return datetime.strptime(date_string, date_format).astimezone(
            pytz.timezone("Europe/Berlin"),
        )
    except (ValueError, TypeError):
        return default



"""Asynchronous Python client providing Urban Data information of Hamburg."""
from __future__ import annotations

import asyncio
import socket
from dataclasses import dataclass
from importlib import metadata
from typing import Any, Self, cast

from aiohttp import ClientError, ClientSession
from aiohttp.hdrs import METH_GET
from yarl import URL

from .exceptions import UDPHamburgConnectionError, UDPHamburgError
from .models import DisabledParking, Garage, ParkAndRide


@dataclass
class UDPHamburg:
    """Main class for handling data fetching from Urban Data Platform of Hamburg."""

    request_timeout: float = 10.0
    session: ClientSession | None = None

    _close_session: bool = False

    async def _request(
        self,
        uri: str,
        *,
        method: str = METH_GET,
        params: dict[str, Any] | None = None,
    ) -> Any:
        """Handle a request to the Urban Data Platform API of Hamburg.

        Args:
        ----
            uri: Request URI, without '/', for example, 'status'
            method: HTTP method to use, for example, 'GET'
            params: Extra options to improve or limit the response.

        Returns:
        -------
            A Python dictionary (text) with the response from
            the Urban Data Platform API.

        Raises:
        ------
            UDPHamburgConnectionError: Timeout occurred while
                connecting to the Urban Data Platform API.
            UDPHamburgError: If the data is not valid.
        """
        version = metadata.version(__package__)
        url = URL.build(
            scheme="https",
            host="api.hamburg.de",
            path="/datasets/v1/",
        ).join(URL(uri))

        headers = {
            "Accept": "application/geo+json",
            "User-Agent": f"PythonUDPHamburg/{version}",
        }

        if self.session is None:
            self.session = ClientSession()
            self._close_session = True

        try:
            async with asyncio.timeout(self.request_timeout):
                response = await self.session.request(
                    method,
                    url,
                    params=params,
                    headers=headers,
                    ssl=True,
                )
                response.raise_for_status()
        except asyncio.TimeoutError as exception:
            msg = "Timeout occurred while connecting to the Urban Data Platform API."
            raise UDPHamburgConnectionError(
                msg,
            ) from exception
        except (ClientError, socket.gaierror) as exception:
            msg = "Error occurred while communicating with Urban Data Platform API."
            raise UDPHamburgConnectionError(
                msg,
            ) from exception

        content_type = response.headers.get("Content-Type", "")
        if "application/geo+json" not in content_type:
            text = await response.text()
            msg = "Unexpected content type response from the Urban Data Platform API"
            raise UDPHamburgError(
                msg,
                {"Content-Type": content_type, "Response": text},
            )

        return cast(dict[str, Any], await response.json())

    async def disabled_parkings(
        self,
        limit: int = 10,
        bulk: str = "false",
    ) -> list[DisabledParking]:
        """Get all disabled parking spaces.

        Args:
        ----
            limit: Number of items to return.
            bulk: Whether to return all items or the limit.

        Returns:
        -------
            A list of DisabledParking objects.
        """
        locations = await self._request(
            "behindertenstellplaetze/collections/verkehr_behindertenparkpl/items",
            params={"limit": limit, "bulk": bulk},
        )
        return [DisabledParking.from_dict(item) for item in locations["features"]]

    async def park_and_rides(
        self,
        limit: int = 10,
        bulk: str = "false",
    ) -> list[ParkAndRide]:
        """Get all park and ride spaces.

        Args:
        ----
            limit: Number of items to return.
            bulk: Whether to return all items or the limit.

        Returns:
        -------
            A list of ParkAndRide objects.
        """
        locations = await self._request(
            "p_und_r/collections/p_und_r/items",
            params={"limit": limit, "bulk": bulk},
        )
        return [ParkAndRide.from_dict(item) for item in locations["features"]]

    async def garages(
        self,
        limit: int = 10,
        bulk: str = "false",
        available: str | None = None,
    ) -> list[Garage]:
        """Get all garages.

        Args:
        ----
            limit: Number of items to return.
            bulk: Whether to return all items or the limit.
            available: Filter based on availability with operators.

        Returns:
        -------
            A list of Garage objects.
        """
        params: dict[str, Any] = {"limit": limit, "bulk": bulk}

        if available is not None:
            params["frei"] = str(available)

        locations = await self._request(
            "parkhaeuser/collections/verkehr_parkhaeuser/items",
            params=params,
        )

        # By default filter out garages without location coordinates.
        return [
            Garage.from_dict(item)
            for item in locations["features"]
            if item["geometry"] is not None
        ]

    async def close(self) -> None:
        """Close open client session."""
        if self.session and self._close_session:
            await self.session.close()

    async def __aenter__(self) -> Self:
        """Async enter.

        Returns
        -------
            The Urban Data Platform object.
        """
        return self

    async def __aexit__(self, *_exc_info: object) -> None:
        """Async exit.

        Args:
        ----
            _exc_info: Exec type.
        """
        await self.close()
